unit Unit1;

interface

{
 Principal Component Analysis
 Dimensionality reduction
 EDUCATIONAL TOOL
 2D

 Ver. 0.2

 Authors:
  Martin Pokorny
  Jana Ludolph

 Based on:
  Lindsay I Smith -- A tutorial on Principal Components Analysis. 2002.
  http://csnet.otago.ac.nz/cosc453/student_tutorials/principal_components.pdf
}  

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, TeeProcs, TeEngine, Chart, XPMan, Series, Math,
  ArrowCha, Unit2;

const
  LARGE_DATASET = 100;                              // Affects point size in graph, number of points in CalcBox
  POINT_SIZE_SMALL = 2;                             // Point size for small data sets
  POINT_SIZE_LARGE = 4;                             // The same for large data sets
  
type
  TSource = String;                                 // Path + filename, data in source file
  
  TDimension = (X, Y);                              // 2D coordinates
  TValue = Real;                                    // Data vector values
  TVectors = array of array[TDimension] of TValue;  // Data vectors (vector 0 = [x, y], vector 1 = [x, y], ..., vector Count - 1 = [x, y])
  TCount = Cardinal;                                // Number of data vectors/points
  TData = record
            Vectors: TVectors;                      // Original vectors
            VectorsAdj: TVectors;                   // Adjusted vectors (memory can be saved here)
            Count: TCount;                          // Number of data vectors/points
            VectorsInMemory: Boolean;               // Flag if data from file is in memory
          end;

  TCovarianceMatrix = array[TDimension, TDimension] of Real;  // Covariance matrix
  TPrincipalComponent = array[TDimension] of TValue;          // Eigenvector with the highest eigenvalue
  
  TStep = (ReadDataStep, SubtractMeanStep, CovarMatrixStep, EigenVvStep, PrincipComponentStep, DerivingNewDataSetStep);
                                                    // Application step according to Smith material

  TProcessState = (Calculating, Finished);          // Flags showing the process state                                                    

  TForm1 = class(TForm)
    btnBack: TButton;
    btnNext: TButton;
    chrtGraph: TChart;
    gbxHelpBox: TGroupBox;
    gbxCalcBox: TGroupBox;
    lblStep: TLabel;
    lblStepName: TLabel;
    btnSaveCalc: TButton;
    btnSaveGraph: TButton;
    XPManifest1: TXPManifest;
    btnData: TButton;
    lblStepNumber: TLabel;
    dlgOpenData: TOpenDialog;
    Series1: TPointSeries;
    memoCalcBox: TMemo;
    lblSourceFile: TLabel;
    dlgSaveGraph: TSaveDialog;
    dlgSaveCalcBox: TSaveDialog;
    lblHelpLine1: TLabel;
    lblHelpLine3: TLabel;
    lblHelpLine2: TLabel;
    lblStepTotalSteps: TLabel;
    lblProcessInfo: TLabel;
    Series2: TLineSeries;
    Series3: TPointSeries;
    Series4: TArrowSeries;
    chbxShowProjectionLines: TCheckBox;
    imgHelpStep3: TImage;
    imgHelpStep4: TImage;
    btnAbout: TButton;
    lblHelpLine4: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);    
    procedure btnDataClick(Sender: TObject);
    procedure btnNextClick(Sender: TObject);
    procedure btnBackClick(Sender: TObject);
    procedure btnSaveCalcClick(Sender: TObject);
    procedure btnSaveGraphClick(Sender: TObject);
    procedure btnAboutClick(Sender: TObject);
    procedure chbxShowProjectionLinesClick(Sender: TObject);
    procedure chrtGraphResize(Sender: TObject);
    procedure FormConstrainedResize(Sender: TObject; var MinWidth, MinHeight, MaxWidth, MaxHeight: Integer);

  private
    { Private declarations }

    // Gui data
    Step: TStep;
    ProcessState: TProcessState;
    StepResult: Boolean;
    FormShown: Boolean;

    // Application logic data
    Source: TSource;
    Data: TData;
    CovarianceMatrix: TCovarianceMatrix;
    PrincipalComponent: TPrincipalComponent;

    // Gui methods
    procedure InitializeGui;
    procedure SetStepNumberAndName;
    procedure SetHelpBox;
    procedure SetNavigationBeforeCalculating;
    procedure SetNavigationAfterCalculating;
    procedure ShowProcessState;
    procedure SetSeries;
    procedure SetPointSize(_Size: Cardinal);
    procedure SetProjectionControl;
    procedure RunStep;

    // Application logic methods
    procedure Initialize;
    procedure ApplicationError(_Message: String);

    procedure ShowPrincipalComponent;
    procedure AdjustMinMaxAxis;
    procedure MoveToProjectionPoints;

    function ReadData: Boolean;
    function SubtractMean: Boolean;
    function CovarMatrix: Boolean;
    function EigenVv: Boolean;
    function PrincipComponent: Boolean;
    function DerivingNewDataSet: Boolean;

  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

// Private type definitions for EigenVv function 
type
  TNumberOfEigenVectors = 0..2;   // 0 no eigenvectors, 1 one eigenvector, 2 eigenvectors
  TEigenvectors = (Vec1, Vec2);
  TEigenvalues = (Val1, Val2);
  TListOfEigenVectors = array[TEigenVectors, TDimension] of TValue;
  TListOfEigenValues = array[TEigenValues] of TValue;
  TEigenVectorsValues = record
                          Number: TNumberOfEigenVectors;
                          Vectors: TListOfEigenVectors;
                          Values: TListOfEigenValues;
                        end;


// ************************* EVENT HANDLERS *******************************

procedure TForm1.FormCreate(Sender: TObject);
begin
  FormShown := False;
  Initialize;
end;

procedure TForm1.FormShow(Sender: TObject);
begin
  if not FormShown then
    InitializeGui;
  FormShown := True;
end;

procedure TForm1.btnDataClick(Sender: TObject);
begin
  if DlgOpenData.Execute then
    begin
      Initialize;
      InitializeGui;
      Source := DlgOpenData.FileName;
      lblSourceFile.Caption := ExtractFileName(Source);
      RunStep;
    end;
end;

procedure TForm1.btnNextClick(Sender: TObject);
begin
  Inc(Step);
  RunStep;
end;

procedure TForm1.btnBackClick(Sender: TObject);
begin
  Dec(Step);
  RunStep;
end;

procedure TForm1.btnSaveCalcClick(Sender: TObject);
begin
  if dlgSaveCalcBox.Execute then
    memoCalcBox.Lines.SaveToFile(dlgSaveCalcBox.FileName);
end;

procedure TForm1.btnSaveGraphClick(Sender: TObject);
begin
  if dlgSaveGraph.Execute then
    chrtGraph.SaveToBitmapFile(dlgSaveGraph.FileName);
end;

procedure TForm1.btnAboutClick(Sender: TObject);
begin
  Form2.ShowModal;
end;

// Obsolete
procedure TForm1.chbxShowProjectionLinesClick(Sender: TObject);
begin
  Series4.Active := chbxShowProjectionLines.Checked;
end;

// The graph should be square to achieve orthogonal projection lines there
procedure TForm1.chrtGraphResize(Sender: TObject);
begin
  chrtGraph.Width := chrtGraph.Height;
end;

// The window size affects the graph size (width), see previous comment
procedure TForm1.FormConstrainedResize(Sender: TObject; var MinWidth, MinHeight, MaxWidth, MaxHeight: Integer);
begin
  MinWidth := chrtGraph.Left + chrtGraph.Width + 20;
  MaxWidth := MinWidth;
end;

// ************************** OWN METHODS ********************************

// ################################# GUI ##################################

procedure TForm1.InitializeGui;
begin
  Step := Low(TStep);
  StepResult := False;
  ProcessState := Finished;

  lblSourceFile.Caption := 'no file';

  Series1.Clear;
  Series2.Clear;
  Series3.Clear;
  Series4.Clear;

  btnData.SetFocus;

  memoCalcBox.Lines.Clear;

  // Obsolete  
  // chbxShowProjectionLInes.Checked := True;

  SetSeries;  

  SetHelpBox;

  SetStepNumberAndName;

  SetNavigationBeforeCalculating;

  ShowProcessState;

  // Obsolete
  // SetProjectionControl;
end;

procedure TForm1.SetStepNumberAndName;
begin
  lblStepNumber.Caption := IntToStr(Ord(Step) + 1);
  case Step of
    ReadDataStep:           lblStepName.Caption := 'Read data set';
    SubtractMeanStep:       lblStepName.Caption := 'Subtract the mean';
    CovarMatrixStep:        lblStepName.Caption := 'Covariance matrix';
    EigenVvStep:            lblStepName.Caption := 'Eigenvectors/eigenvalues';
    PrincipComponentStep:   lblStepName.Caption := 'Principal component';
    DerivingNewDataSetStep: lblStepName.Caption := 'Deriving new data set';
  end;
end;

procedure TForm1.SetHelpBox;
begin
  lblHelpLine1.Visible := (Step <> CovarMatrixStep) and (Step <> EigenVvStep);
  lblHelpLine2.Visible := LblHelpLine1.Visible;
  lblHelpLine3.Visible := LblHelpLine1.Visible;
  lblHelpLine4.Visible := LblHelpLine1.Visible;  

  imgHelpStep3.Visible := Step = CovarMatrixStep;
  imgHelpStep4.Visible := Step = EigenVvStep;

  case Step of
    ReadDataStep:
      begin
        lblHelpLine1.Caption := '- Open file with data vectors';
        lblHelpLine2.Caption := '- Each data vector per line';
        lblHelpLine3.Caption := '- Dimensions separated by space/tab';
        lblHelpLine4.Caption := '- Graph: zoom with LMB, shift with RMB';
      end;
    SubtractMeanStep:
      begin
        lblHelpLine1.Caption := '- Calculate mean for each dimension';
        lblHelpLine2.Caption := '- Subtract the mean for each vector';
        lblHelpLine3.Caption := '- Get new data set with zero mean';
        lblHelpLine4.Caption := '';
      end;
    PrincipComponentStep:
      begin
        lblHelpLine1.Caption := 'Principal component is the eigenvector with';
        lblHelpLine2.Caption := 'the highest eigenvalue. It is the most signifi-';
        lblHelpLine3.Caption := 'cant relationship between the data dimensions.';
        lblHelpLine4.Caption := 'Dimensionality reduction: see next step.';
      end;
    DerivingNewDataSetStep:
      begin
        lblHelpLine1.Caption := '- Multiply PC vector by data vectors';
        lblHelpLine2.Caption := '- Data expressed in terms of PC';
        lblHelpLine3.Caption := '- Principal component becoms the X-axis';
        lblHelpLine4.Caption := '- For details see About reference';
      end;
  end;
end;

// Large data sets require buttons to be set before the calculation
procedure TForm1.SetNavigationBeforeCalculating;
begin
  btnBack.Enabled := Step <> Low(TStep);
  btnNext.Enabled := (Step <> High(TStep)) and Data.VectorsInMemory;
end;

// Depends on step result
procedure TForm1.SetNavigationAfterCalculating;
begin
  if (Step = ReadDataStep) or (Step = EigenVvStep) then
    btnNext.Enabled := StepResult
end;

procedure TForm1.ShowProcessState;
begin
  if Step = ReadDataStep then
    lblProcessInfo.Caption := 'Reading data...'
  else
    lblProcessInfo.Caption := 'Calculating...';

  lblProcessInfo.Visible := ProcessState = Calculating;
  memoCalcBox.Visible := not lblProcessInfo.Visible;
end;

procedure TForm1.SetSeries;
begin
  // For large dataset to avoid blinking
  if Step = PrincipComponentStep then
    begin
      Series2.Clear;
      Series3.Clear;
      Series4.Clear;
      chrtGraph.Refresh;
    end;

  Series1.Active := Step <> DerivingNewDataSetStep;
  Series2.Active := (Step = PrincipComponentStep) or (Step = DerivingNewDataSetStep);
  Series3.Active := Series2.Active;
  // Obsolete  
  // Series4.Active := (Step = PrincipComponentStep) and chbxShowProjectionLines.Checked;
  Series4.Active := Step = PrincipComponentStep;  
end;

procedure TForm1.SetPointSize(_Size: Cardinal);
begin
  Series1.Pointer.HorizSize := _Size;
  Series1.Pointer.VertSize := _Size;
  Series3.Pointer.HorizSize := _Size;
  Series3.Pointer.VertSize := _Size;
end;

// Obsolete
procedure TForm1.SetProjectionControl;
begin
  chbxShowProjectionLines.Enabled := Step = PrincipComponentStep;
end;

procedure TForm1.RunStep;
begin
  // Application improvement possibility:
  // CPU time can be saved -- mark already finished
  // computations, save results to object data
  // and do not process them again.

  SetNavigationBeforeCalculating;

  SetStepNumberAndName;

  SetHelpBox;

  SetSeries;

  // Obsolete
  // SetProjectionControl;

  ProcessState := Calculating;
  ShowProcessState;

  Form1.Refresh;

  case Step of
    ReadDataStep:           StepResult := ReadData;
    SubtractMeanStep:       StepResult := SubtractMean;
    CovarMatrixStep:        StepResult := CovarMatrix;
    EigenVvStep:            StepResult := EigenVv;
    PrincipComponentStep:   StepResult := PrincipComponent;
    else                    StepResult := DerivingNewDataSet;
  end;

  ProcessState := Finished;
  ShowProcessState;

  SetNavigationAfterCalculating;
end;


// ############################## APP LOGIC ###############################

procedure TForm1.Initialize;
begin
  Source := '';
  Data.Vectors := nil;
  Data.VectorsAdj := nil;
  Data.Count := 0;
  Data.VectorsInMemory := False;
  // Covariance matrix doesn't have to be initialized
  // Principal component doesn't have to be initialized
end;

procedure TForm1.ApplicationError(_Message: String);
begin
  MessageDlg('Processing error:' + #13#10 + _Message + #13#10 + #13#10 + 'Application will be closed.', mtWarning, [mbOk], 0);
  Application.Terminate;
end;

procedure TForm1.ShowPrincipalComponent;
var K: Real; // K = slope
begin
  Series2.Clear;

  if PrincipalComponent[x] = 0 then
    begin
      // K is infinity
      // principal component is the y-axis.
      Series2.AddXY(0, chrtGraph.LeftAxis.Maximum);
      Series2.AddXY(0, chrtGraph.LeftAxis.Minimum);
    end
  else
    begin
      K := PrincipalComponent[y] / PrincipalComponent[x];

      // Increasing
      if K >= 0 then
        // <= 45 , this is also the case for PrincipalComponent[y]=0. That leads to k=0. And x-axis as principal component
        if abs(PrincipalComponent[x]) >= abs(PrincipalComponent[y]) then
          begin
            Series2.AddXY(chrtGraph.BottomAxis.Maximum, K * chrtGraph.BottomAxis.Maximum);
            Series2.AddXY(chrtGraph.BottomAxis.Minimum, K * chrtGraph.BottomAxis.Minimum);
          end
        // > 45 
        else
          begin
            Series2.AddXY(chrtGraph.LeftAxis.Maximum / K, chrtGraph.LeftAxis.Maximum);
            Series2.AddXY(chrtGraph.LeftAxis.Minimum / K, chrtGraph.LeftAxis.Minimum);
          end
      // Decreasing
      else
        // < 135 
        if abs(PrincipalComponent[y]) > abs(PrincipalComponent[x]) then
          begin
            Series2.AddXY(chrtGraph.LeftAxis.Maximum / K, chrtGraph.LeftAxis.Maximum);
            Series2.AddXY(chrtGraph.LeftAxis.Minimum / K, chrtGraph.LeftAxis.Minimum);
          end
        // >= 135 
        else
          begin
            Series2.AddXY(chrtGraph.BottomAxis.Minimum, K * chrtGraph.BottomAxis.Minimum);
            Series2.AddXY(chrtGraph.BottomAxis.Maximum, K * chrtGraph.BottomAxis.Maximum);
          end;
    end;
end;


// This procedure is necessary to achieve orthogonal projection lines to principal component
procedure TForm1.AdjustMinMaxAxis;
var MinX, MaxX, MinY, MaxY, Min, Max, Margin: Double;
begin
  if Series1.Active and Series3.Active then
    begin
      if Series3.MinXValue < Series1.MinXValue then
        MinX := Series3.MinXValue
      else
        MinX := Series1.MinXValue;

      if Series3.MaxXValue > Series1.MaxXValue then
        MaxX := Series3.MaxXValue
      else
        MaxX := Series1.MaxXValue;

      if Series3.MinYValue < Series1.MinYValue then
        MinY := Series3.MinYValue
      else
        MinY := Series1.MinYValue;

      if Series3.MaxYValue > Series1.MaxYValue then
        MaxY := Series3.MaxYValue
      else
        MaxY := Series1.MaxYValue;
    end

  else if Series1.Active then
    begin
      MinX := Series1.MinXValue;
      MinY := Series1.MinYValue;
      MaxX := Series1.MaxXValue;
      MaxY := Series1.MaxYValue;
    end

  else if Series3.Active then
    begin
      MinX := Series3.MinXValue;
      MinY := Series3.MinYValue;
      MaxX := Series3.MaxXValue;
      MaxY := Series3.MaxYValue;
    end

  else
    begin
      MinX := 0;
      MinY := 0;
      MaxX := 0;
      MaxY := 0;
    end;

  if MinX < MinY then
    Min := MinX
  else
    Min := MinY;

  if MaxX > MaxY then
    Max := MaxX
  else
    Max := MaxY;

  chrtGraph.BottomAxis.Automatic := False;
  chrtGraph.LeftAxis.Automatic := False;

  Margin := abs(Max - Min) * 0.05;

  chrtGraph.BottomAxis.SetMinMax(Min - Margin, Max + Margin);
  chrtGraph.LeftAxis.SetMinMax(Min - Margin, Max + Margin);
end;


{ Following functions solves the PCA steps.
  One of optimalization possibilities is
  to store results in each step and do
  not calculate them again in each step.
}

// STEP 1: Read original data set (return value is true only if Data.Count > 0)
function TForm1.ReadData: Boolean;
var F: TextFile;
    i: TCount;
    j: Cardinal;
begin
  Series1.Clear;
  memoCalcBox.Lines.Clear;

  memoCalcBox.Lines.Add('File: ' + ExtractFileName(Source));
  
  for j := 1 to 4 do
    memoCalcBox.Lines.Add('');

  if not Data.VectorsInMemory then
    begin
      Data.Vectors := nil;
      Data.Count := 0;

      AssignFile(F, Source);

      {$I-}
      FileMode := 0;  {Set file access to read only}
      Reset(F);
      {$I+}

      if (IOResult = 0) and (Source <> '') then
        begin
          i := 0;

          while not eof(F) do
            begin
              try
                SetLength(Data.Vectors, i + 1);
              except
                on EOutOfMemory do
                  ApplicationError('Not enough memory to store the data.');
              end;

              try
                ReadLn(F, Data.Vectors[i,X], Data.Vectors[i,Y]);
              except
                on EInOutError do
                  ApplicationError('File read error. Error reading line: ' + IntToStr(i + 1));
              end;

              Series1.AddXY(Data.Vectors[i,X], Data.Vectors[i,Y]);

              if (i + 1) <= LARGE_DATASET then
                memoCalcBox.Lines.Add(FloatToStr(Data.Vectors[i,X]) + '; ' + FloatToStr(Data.Vectors[i,Y]));

              Inc(i);
            end;

          CloseFile(F);

          Data.Count := i;
        end
      else
        ApplicationError('File read error. Can not open file: ' + Source);

      Data.VectorsInMemory := Data.Count > 0;
    end

  else
    for i := 0 to Data.Count - 1 do
      begin
        Series1.AddXY(Data.Vectors[i,X], Data.Vectors[i,Y]);
        if (i + 1) <= LARGE_DATASET then
          memoCalcBox.Lines.Add(FloatToStr(Data.Vectors[i,X]) + '; ' + FloatToStr(Data.Vectors[i,Y]));
      end;

  if Data.Count < LARGE_DATASET then
    SetPointSize(4)
  else
    SetPointSize(2);

  if Data.Count > 0 then
    begin
      AdjustMinMaxAxis;
      memoCalcBox.Lines[2] := IntToStr(Data.Count) + ' data vectors:';
      memoCalcBox.Lines[3] := '(shown first ' + IntToStr(LARGE_DATASET) + ' points)';
    end;

  Result := Data.Count > 0;
end;


// STEP 2: Subtract the mean for each dimension
function TForm1.SubtractMean: Boolean;
var Dim: TDimension;
    Sum: Real;
    i: TCount;
    Mean: Real;
    RoundedVal1, RoundedVal2: String;
begin
  Data.VectorsAdj := nil;

  try
    SetLength(Data.VectorsAdj, Data.Count);
  except
    on EOutOfMemory do
      ApplicationError('Not enough memory to store the data.');
  end;

  Series1.Clear;  
  memoCalcBox.Lines.Clear;

  memoCalcBox.Lines.Add('Dimension mean:');   

  for Dim := Low(TDimension) to High(TDimension) do
    begin
      Sum := 0;

      for i := 0 to Data.Count - 1 do
        Sum := Sum + Data.Vectors[i,Dim];

      // Data.Count is greater then zero
      if Data.Count > 0 then
        Mean := Sum / Data.Count
      else
        Mean := 0;

      Str(Mean:0:3, RoundedVal1);
      if Dim = X then
        memoCalcBox.Lines.Add('X dimension: ' + RoundedVal1)
      else
        memoCalcBox.Lines.Add('Y dimension: ' + RoundedVal1);

      for i := 0 to Data.Count - 1 do
        Data.VectorsAdj[i,Dim] := Data.Vectors[i,Dim] - Mean;
    end;

  memoCalcBox.Lines.Add('');    
  memoCalcBox.Lines.Add('Data set with zero mean:');
  memoCalcBox.Lines.Add('(shown first ' + IntToStr(LARGE_DATASET) + ' points)');
  memoCalcBox.Lines.Add('');

  for i := 0 to Data.Count - 1 do
    begin
      Series1.AddXY(Data.VectorsAdj[i,X], Data.VectorsAdj[i,Y]);
      if (i + 1) <= LARGE_DATASET then
        begin
          Str(Data.VectorsAdj[i,X]:0:3, RoundedVal1);
          Str(Data.VectorsAdj[i,Y]:0:3, RoundedVal2);
          memoCalcBox.Lines.Add(RoundedVal1 + '; ' + RoundedVal2);
        end;
    end;

  AdjustMinMaxAxis;

  Result := True;
end;


// STEP 3: Covariance matrix calculation
function TForm1.CovarMatrix: Boolean;
var i: TCount;
    SumX, SumY: Real;
    MeanX, MeanY: Real;
    RoundedVal1, RoundedVal2 : String;
begin
  memoCalcBox.Lines.Clear;

  SumX := 0;
  SumY := 0;

  for i := 0 to Data.Count - 1 do
    begin
      SumX := SumX + Data.VectorsAdj[i,X];
      SumY := SumY + Data.VectorsAdj[i,Y];
    end;

  // Data.Count is grater then zero
  if Data.Count > 0 then
    begin
      MeanX := SumX / Data.Count;
      MeanY := SumY / Data.Count;
    end
  else
    begin
      MeanX := 0;
      MeanY := 0;
    end;

  CovarianceMatrix[X,X] := 0;
  CovarianceMatrix[X,Y] := 0;
  CovarianceMatrix[Y,Y] := 0;

  for i := 0 to Data.Count - 1 do
    begin
      CovarianceMatrix[X,X] := CovarianceMatrix[X,X] + ( (Data.VectorsAdj[i,X] - MeanX) * (Data.VectorsAdj[i,X] - MeanX) );
      CovarianceMatrix[X,Y] := CovarianceMatrix[X,Y] + ( (Data.VectorsAdj[i,X] - MeanX) * (Data.VectorsAdj[i,Y] - MeanY) );
      CovarianceMatrix[Y,Y] := CovarianceMatrix[Y,Y] + ( (Data.VectorsAdj[i,Y] - MeanY) * (Data.VectorsAdj[i,Y] - MeanY) );
    end;

  // Data.Count is grater then zero
  if Data.Count > 0 then
    begin
      CovarianceMatrix[X,X] := CovarianceMatrix[X,X] / Data.Count;
      CovarianceMatrix[X,Y] := CovarianceMatrix[X,Y] / Data.Count;
      CovarianceMatrix[Y,Y] := CovarianceMatrix[Y,Y] / Data.Count;
    end;

  CovarianceMatrix[Y,X] := CovarianceMatrix[X,Y];

  memoCalcBox.Lines.Add('Covariance matrix:');
  memoCalcBox.Lines.Add('');

  Str(CovarianceMatrix[X,X]:0:3, RoundedVal1);
  Str(CovarianceMatrix[X,Y]:0:3, RoundedVal2);
  memoCalcBox.Lines.Add(RoundedVal1 + '    ' + RoundedVal2);

  Str(CovarianceMatrix[Y,X]:0:3, RoundedVal1);
  Str(CovarianceMatrix[Y,Y]:0:3, RoundedVal2);
  memoCalcBox.Lines.Add(RoundedVal1 + '    ' + RoundedVal2);

  Result := True;
end;


// STEP 4: Eigenvalues and eigenvectors calculation
function TForm1.EigenVv: Boolean;
var EigenVectorsValues: TEigenVectorsValues;
    D, VLength: Real;
    EVecVal : TEigenValues;
    RoundedVal1, RoundedVal2: String;
begin
  memoCalcBox.Lines.Clear;

  memoCalcBox.Lines.Add('Eigenvectors/eigenvalues for the covariance matrix:');
  memoCalcBox.Lines.Add('');

{
  Used method: symbolic computation

  In smaller matrices the eigenvalues can be calculated with the
  characteristic polynom. The characteristic polynom of a matrix A is det(A - delta*I) = 0,
  where delta is an eigevalue and I is the identity matrix. In
  case of 2d covariance matrix, the characteristic polynom would be:
  delta^2 + (-cov(xx) -cov(yy))delta + (cov(xx)*cov(yy) - cov(xy)*cov(yx)). We have to
  calculate delta.
  The related determinante is:  ((-cov(xx)-cov(yy))/2)^2 - (cov(xx)*cov(yy) - cov(xy)*cov(yx))
}

  D:= ((-CovarianceMatrix[X,X] - CovarianceMatrix[Y,Y]) * (-CovarianceMatrix[X,X] - CovarianceMatrix[Y,Y])) / 4 - (CovarianceMatrix[X,X] * CovarianceMatrix[Y,Y] - CovarianceMatrix[X,Y] * CovarianceMatrix[Y,X]);

  if D > 0 then
    with EigenVectorsValues do
      begin
	      Values[Val1] := (CovarianceMatrix[X,X] + CovarianceMatrix[Y,Y]) / 2 + sqrt(D);
	      Values[Val2] := (CovarianceMatrix[X,X] + CovarianceMatrix[Y,Y]) / 2 - sqrt(D);
        Str(Values[Val1]:0:3, RoundedVal1);
        memoCalcBox.Lines.Add('Eigenvalue 1 = ' + RoundedVal1);
        Str(Values[Val2]:0:3, RoundedVal2);
        memoCalcBox.Lines.Add('Eigenvalue 2 = ' + RoundedVal2);
        memoCalcBox.Lines.Add('');
        Number := 2;
        for EVecVal := Val1 to Val2 do
          begin
            	// To calculate the eigenvector (x,y) according to the eigenvalue e, we have the following linear equation system:
          		// (CovarianceMatrix[X,X] - e)*x + CovarianceMatrix[X,Y]*y = 0
          		// CovarianceMatrix[Y,X]*x + (CovarianceMatrix[Y,Y] - e)*y = 0
          		// There are now 4 ways of getting a result (beside the trivial case, when x=y=0).
          		// 1. We start with x=1
          		// 1.a. then y = -(CovarianceMatrix[X,X] - e ) / CovarianceMatrix[X,y]
          		// 1.b  . or y = -CovarianceMatrix[y,x] / (CovarianceMatrix[y,y] - e)
              // We chose one of them, if we don't have to divide to zero.
          		// 2. We start with y=1
          		// 2.a. then x = - (CovarianceMatrix[y,y] - e) / CovarianceMatrix[y,x]
          		// 2.b. or x = - CovarianceMatrix[x,y] / (CovarianceMatrix[x,x] - e)
          		// When the equation system looks like
          		// 0*x + y = 0
          		// 0*x + y = 0
          		// We have only the trivial case (x,y) = (0,0)
              // This case is defined as a non-valid one, since it is no helpful principal axis.

      	    if (CovarianceMatrix[X,Y] <> 0) then
          		begin
                	Vectors[TEigenVectors(EVecVal),x] := 1;
                	Vectors[TEigenVectors(EVecVal),y] := -(CovarianceMatrix[X,X] - Values[EVecVal] ) / CovarianceMatrix[X,y];
          		end
	          else if ((CovarianceMatrix[Y,Y]- Values[EVecVal]) <> 0) then
          		begin
          			  Vectors[TEigenVectors(EVecVal),x] := 1;
                	Vectors[TEigenVectors(EVecVal),y] := -CovarianceMatrix[y,x] / (CovarianceMatrix[y,y] - Values[EVecVal]);
          		end
      	    else if (CovarianceMatrix[Y,X] <> 0) then
		          begin
                	Vectors[TEigenVectors(EVecVal),y] := 1;
                	Vectors[TEigenVectors(EVecVal),x] := - (CovarianceMatrix[y,y] - Values[EVecVal]) / CovarianceMatrix[y,x];
          		end
      	    else if ((CovarianceMatrix[X,X]-Values[EVecVal]) <> 0) then
          		begin
            			Vectors[TEigenVectors(EVecVal),y] := 1;
                	Vectors[TEigenVectors(EVecVal),x] := - CovarianceMatrix[x,y] / (CovarianceMatrix[x,x] - Values[EVecVal]);
          		end
      	    else
          		begin
      			      // When all values are zero, then the eigenvector would be (0,0), it is no valid one.
			            Dec(Number);
			            Vectors[TEigenVectors(EVecVal),x] := 0;
                	Vectors[TEigenVectors(EVecVal),y] := 0;
          		end;

	          // This branch is now to check if there is a valid eigenvector, which should have at least one non-zero value, or not. The same for case D=0.
	          if ((Vectors[TEigenVectors(EVecVal),x] = 0) and (Vectors[TEigenVectors(EVecVal),y] = 0)) then
		          begin
			          memoCalcBox.Lines.Add('Calculation problems occured:');
                memoCalcBox.Lines.Add('It is not possible to calculate the eigenvector ' + IntToStr(Ord(EVecVal) + 1) + '.');
		          end
      	    else
		          begin
			          //now here it is ensured, that the Length of the vector can't be zero.
			          VLength := sqrt(Power(Vectors[TEigenVectors(EVecVal),x], 2) + Power(Vectors[TEigenVectors(EVecVal),y], 2));
	            	Vectors[TEigenVectors(EVecVal),x] := Vectors[TEigenVectors(EVecVal),x] / VLength;
        		    Vectors[TEigenVectors(EVecVal),y] := Vectors[TEigenVectors(EVecVal),y] / VLength;
           		  Str(Vectors[TEigenVectors(EVecVal),x]:0:3, RoundedVal1);
            		Str(Vectors[TEigenVectors(EVecVal),y]:0:3, RoundedVal2);
            		memoCalcBox.Lines.Add('Eigenvector ' + IntToStr(Ord(EVecVal) + 1) + ' = (' + RoundedVal1 + '; ' + RoundedVal2 + ')');
		          end;
          end;
      end

  else if D = 0 then
    with EigenVectorsValues do
      begin
        Values[Val1] := (CovarianceMatrix[X,X] + CovarianceMatrix[Y,Y]) / 2;
        Values[Val2] := Values[Val1];
        Str(Values[Val1]:0:3, RoundedVal1);
        memoCalcBox.Lines.Add('Eigenvalue 1 = ' + RoundedVal1);
        memoCalcBox.Lines.Add('Eigenvalue 2 = the same as eigenvalue 1');
        memoCalcBox.Lines.Add('');
        Number := 1;

	      if (CovarianceMatrix[X,Y] <> 0) then
		      begin
			      Vectors[Vec1,x] := 1;
            Vectors[Vec1,y] := -( CovarianceMatrix[X,X] - Values[Val1] ) / CovarianceMatrix[X,y];
		      end
	      else if ((CovarianceMatrix[Y,Y]- Values[Val1]) <> 0) then
		      begin
            Vectors[Vec1,x] := 1;
            Vectors[Vec1,y] := -CovarianceMatrix[y,x] / (CovarianceMatrix[y,y] - Values[Val1]);
		      end
	      else if (CovarianceMatrix[Y,X] <> 0) then
		      begin
			      Vectors[Vec1,y] := 1;
            Vectors[Vec1,x] := - (CovarianceMatrix[y,y] - Values[Val1]) / CovarianceMatrix[y,x];
		      end
	      else if ((CovarianceMatrix[X,X]-Values[Val1]) <> 0) then
		      begin
			      Vectors[Vec1,y] := 1;
            Vectors[Vec1,x] := - CovarianceMatrix[x,y] / (CovarianceMatrix[x,x] - Values[Val1]);
		      end
	      else
		      begin
			      Number := 0;
        		Vectors[Vec1,x] := 0;
        		Vectors[Vec1,y] := 0;
        		Vectors[Vec2,x] := 0;
        		Vectors[Vec2,y] := 0;
		      end;

	      if ((Vectors[Vec1,x] = 0) and (Vectors[Vec1,y] = 0)) then
		      begin
			      memoCalcBox.Lines.Add('Calculation problems occured:');
      			memoCalcBox.Lines.Add('It is not possible to calculate the eigenvector.');
		      end
	      else
		      begin
			      VLength := sqrt(Power(Vectors[Vec1,x], 2) + Power(Vectors[Vec1,y], 2));
	        	Vectors[Vec1,x] := Vectors[Vec1,x] / VLength;
        		Vectors[Vec1,y] := Vectors[Vec1,y] / VLength;
        		Vectors[Vec2,x] := Vectors[Vec1,x];
        		Vectors[Vec2,y] := Vectors[Vec1,y];
        		Str(Vectors[Vec1,x]:0:3, RoundedVal1);
        		Str(Vectors[Vec1,y]:0:3, RoundedVal2);
        		memoCalcBox.Lines.Add('Eigenvector 1 = (' + RoundedVal1 + '; ' + RoundedVal2 + ')');
        		memoCalcBox.Lines.Add('Eigenvector 2: the same as eigenvector 1');
		      end;
      end

  else //in case D<0, the characteristic polynom has no results, therefor no eigenvalues nor eigenvectors
    with EigenVectorsValues do
      begin
        Number := 0;
        Values[Val1] := 0;
        Values[Val2] := 0;
        Vectors[Vec1,x] := 0;
        Vectors[Vec1,y] := 0;
        Vectors[Vec2,x] := 0;
        Vectors[Vec2,y] := 0;
        memoCalcBox.Lines.Add('Calculation problems occured:');
        memoCalcBox.Lines.Add('It is not possible to calculate both of the eigenvalues and eigenvectors.');
      end;

  if EigenVectorsValues.Number = 2 then
    if EigenVectorsValues.Values[Val1] >= EigenVectorsValues.Values[Val2] then
      begin
        PrincipalComponent[x] := EigenVectorsValues.Vectors[Vec1, x];
        PrincipalComponent[y] := EigenVectorsValues.Vectors[Vec1, y];
      end
    else
      begin
        PrincipalComponent[x] := EigenVectorsValues.Vectors[Vec2, x];
        PrincipalComponent[y] := EigenVectorsValues.Vectors[Vec2, y];
      end
  else if EigenVectorsValues.Number = 1 then
    begin
      PrincipalComponent[x] := EigenVectorsValues.Vectors[Vec1, x];
      PrincipalComponent[y] := EigenVectorsValues.Vectors[Vec1, y];
    end
  else
    begin
      PrincipalComponent[x] := 0;
      PrincipalComponent[y] := 0;
    end;

  Result := EigenVectorsValues.Number <> 0;
end;


// STEP 5: Calculate final data set, plot principal component and projection points on the principal component line
function TForm1.PrincipComponent: Boolean;
var i: TCount;
    Projection, ProjectionPointX, ProjectionPointY: Real;
    RoundedValue1, RoundedValue2: String;
begin
  Series2.Clear;
  Series3.Clear;
  Series4.Clear;
  chrtGraph.Refresh;

  with memoCalcBox.Lines do
    begin
      Clear;
      Add('Principal component:');
      Str(PrincipalComponent[X]:0:3, RoundedValue1);
      Str(PrincipalComponent[Y]:0:3, RoundedValue2);
      Add('red dashed line, vector = (' + RoundedValue1 + '; ' + RoundedValue2 + ')');
      Add('');
      Add('New data set:');
      Add('(shown first ' + IntToStr(LARGE_DATASET) + ' points)');
      Add('');
    end;

  for i := 0 to Data.Count - 1 do
    begin
      Projection := PrincipalComponent[X] * Data.VectorsAdj[i,X] + PrincipalComponent[Y] * Data.VectorsAdj[i,Y];

      if (i + 1) <= LARGE_DATASET then
        begin
          Str(Projection:0:3, RoundedValue1);
          memoCalcBox.Lines.Add(RoundedValue1);
        end;

      ProjectionPointX := PrincipalComponent[X] * Projection;
      ProjectionPointY := PrincipalComponent[Y] * Projection;

      Series4.AddArrow(Data.VectorsAdj[i,X], Data.VectorsAdj[i,Y], ProjectionPointX, ProjectionPointY);
      Series3.AddXY(ProjectionPointX, ProjectionPointY);
    end;

  AdjustMinMaxAxis;

  ShowPrincipalComponent;

  // Improvement comment #1
  chrtGraph.Refresh;
  Sleep(1000);
  Series4.Active := False;
  MoveToProjectionPoints;  

  Result := True;
end;


// STEP 6: Rotate final data set to X axis, principal component line becomes the X axis
function TForm1.DerivingNewDataSet: Boolean;
var AngleDeg, ShiftAngleDeg, Projection: Real;
    Shifts, s, p: Cardinal;
begin
  memoCalcBox.Lines[1] := 'red dashed line is the X-axis.';

  chrtGraph.BottomAxis.Automatic := True;
  chrtGraph.LeftAxis.Automatic := True;

  // Angle between principal component line and X axis
  // In degrees, Angle = (-90, 90) degrees
  // So, it's always the smaller angle, positive in 1st quadrant, negative in 2nd quadrant
  if PrincipalComponent[X] = 0 then
    // When PrincipalComponent[x] = 0 then we have the y axis as pc. Here the angle is 90 degrees
    AngleDeg := 90
  else
    // When PrincipalComponent[y] = 0 then we have the x axis as pc. Here the angle is 0 degrees
    AngleDeg := (ArcTan(PrincipalComponent[Y] / PrincipalComponent[X]) * 180) / PI;

  // Elementary shift angle (step) in degrees, one degree
  ShiftAngleDeg := 1;

  // How many rotation shifts (= whole degrees) must be performed
  Shifts := Trunc(Abs(AngleDeg)/ShiftAngleDeg);

  // Repeat Shift-times if there is anything to rotate
  if Shifts > 0 then
    for s := 1 to Shifts do
      begin
        // Rotation:
        // If the principal component line is increasing (AngleDeg is positive), then always rotate to the right, if it is decreasing, rotate to the left;
        // (*) In case of principal component eigenvector being in the oposite direction to the principal component line slope, reverse points around the origin.

        // AngleDeg is decreased in every shift, its sign must be preserved
        if AngleDeg > 0 then
          AngleDeg := AngleDeg - ShiftAngleDeg
        else
          AngleDeg := AngleDeg + ShiftAngleDeg;

        // Replot all projection points (which are in Series3)
        for p := 0 to Series3.Count - 1 do
          begin
            // Projection can be precalculated in previous step and CPU time can be saved here
            // Projection = radius for the point on the rotation circle, sign is preserved
            // Projection sign > 0 if the point is in the direction of principal component eigenvector, < 0 otherwise
            Projection := PrincipalComponent[X] * Data.VectorsAdj[p,X] + PrincipalComponent[Y] * Data.VectorsAdj[p,Y];
            // Projection, projection sign and equations below define new point coordinates in all four quadrants,
            // in case of (*) points are reversed around the origin.
            Series3.XValues[p] := Projection * Cos(Abs((AngleDeg * PI)/180));
            Series3.YValues[p] := Projection * Sin((AngleDeg * PI)/180);
          end;

        // Refresh principal component line and moved points
        ShowPrincipalComponent;
        chrtGraph.Refresh;

        // Sleep 30 ms experimentaly gives good video
        Sleep(30);
      end;

  // Plot points to their final position
  for p := 0 to Series3.Count - 1 do
    begin
      // Projection can be precalculated in previous step and CPU time can be saved here
      Projection := PrincipalComponent[X] * Data.VectorsAdj[p,X] + PrincipalComponent[Y] * Data.VectorsAdj[p,Y];
      Series3.XValues[p] := Projection;
      Series3.YValues[p] := 0;
    end;

  // Plot Principal component line to the final position
  Series2.Clear;
  Series2.AddXY(chrtGraph.BottomAxis.Minimum, 0);
  Series2.AddXY(chrtGraph.BottomAxis.Maximum, 0);

  Result := True;
end;


// Solves improvement comment #1 -- PART OF STEP 5
// Move adjusted data points to projection points on the principal component line.
// Encountered problems working only with graph series data (different point order
// at least in Series1), so, point coordinates calculated here again (speed can be improved here).
procedure TForm1.MoveToProjectionPoints;
var Shift: Cardinal;
    i: TCount;
    OrigVector, StepVector: array[TDimension] of Real;
    Projection, ProjectionPointX, ProjectionPointY: Real;
begin
  for Shift := 1 to 30 do
    begin
      for i := 0 to Data.Count - 1 do
        begin
        Projection := PrincipalComponent[X] * Data.VectorsAdj[i,X] + PrincipalComponent[Y] * Data.VectorsAdj[i,Y];

        ProjectionPointX := PrincipalComponent[X] * Projection;
        ProjectionPointY := PrincipalComponent[Y] * Projection;

        OrigVector[x] := ProjectionPointX - Data.VectorsAdj[i,X];
        OrigVector[y] := ProjectionPointY - Data.VectorsAdj[i,Y];

        StepVector[x] := OrigVector[x] * (Shift/30);
        StepVector[y] := OrigVector[y] * (Shift/30);

        Series1.XValues[i] := Data.VectorsAdj[i,X] + StepVector[x];
        Series1.YValues[i] := Data.VectorsAdj[i,Y] + StepVector[y];
      end;
      chrtGraph.Refresh;
      Sleep(Round(1000/30));
    end;

  Series1.Active := False;
  Series1.Clear;
  for i := 0 to Data.Count - 1 do
    Series1.AddXY(Data.VectorsAdj[i,X], Data.VectorsAdj[i,Y]);
end;


end.
